@@ -1,24 +1,34 @@
 import contextlib
+from dataclasses import asdict
 from datetime import datetime
-from typing import Any, Callable, ContextManager, Dict, Iterator, List, Optional, Union
+from typing import (
+    Any,
+    Callable,
+    ContextManager,
+    Dict,
+    KeysView,
+    Iterator,
+    List,
+    Optional,
+    Union,
+)
 
+from jinja2 import BaseLoader, Environment
 import pandas as pd
-import pyarrow
 import pyarrow as pa
-import pyarrow.parquet
 from psycopg2 import sql
 from pydantic import StrictStr
 from pydantic.typing import Literal
 
 from feast.data_source import DataSource
-from feast.feature_view import FeatureView
+from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL, FeatureView
 from feast.infra.offline_stores import offline_utils
 from feast.infra.offline_stores.offline_store import OfflineStore, RetrievalJob
 from feast.on_demand_feature_view import OnDemandFeatureView
 from feast.registry import Registry
 from feast.repo_config import RepoConfig
 from feast_postgres.type_map import pg_type_code_to_arrow
-from feast_postgres.utils import _get_conn, df_to_postgres_table, sql_to_postgres_table
+from feast_postgres.utils import _get_conn, df_to_postgres_table, get_query_schema
 
 from ..postgres_config import PostgreSQLConfig
 from .postgres_source import PostgreSQLSource
@@ -65,6 +75,7 @@ class PostgreSQLOfflineStore(OfflineStore):
         query = f"""
             SELECT
                 {b_field_string}
+                {f", {repr(DUMMY_ENTITY_VAL)} AS {DUMMY_ENTITY_ID}" if not join_key_columns else ""}
             FROM (
                 SELECT {a_field_string},
                 ROW_NUMBER() OVER({partition_by_join_key_string} ORDER BY {timestamp_desc_string}) AS _feast_row
@@ -93,8 +104,16 @@ class PostgreSQLOfflineStore(OfflineStore):
     ) -> RetrievalJob:
         @contextlib.contextmanager
         def query_generator() -> Iterator[str]:
-            table_name = offline_utils.get_temp_entity_table_name()
-            entity_schema = _df_to_table(config.offline_store, entity_df, table_name)
+            table_name = None
+            if isinstance(entity_df, pd.DataFrame):
+                table_name = offline_utils.get_temp_entity_table_name()
+                entity_schema = df_to_postgres_table(config.offline_store, entity_df, table_name)
+                df_query = table_name
+            elif isinstance(entity_df, str):
+                df_query = f"({entity_df}) AS sub"
+                entity_schema = get_query_schema(config.offline_store, df_query)
+            else:
+                raise TypeError(entity_df)
 
             entity_df_event_timestamp_col = offline_utils.infer_event_timestamp_from_entity_df(
                 entity_schema
@@ -112,25 +131,25 @@ class PostgreSQLOfflineStore(OfflineStore):
                 feature_refs, feature_views, registry, project,
             )
 
-            query = offline_utils.build_point_in_time_query(
-                query_context,
-                left_table_query_string=table_name,
-                entity_df_event_timestamp_col=entity_df_event_timestamp_col,
-                query_template=MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN,
-                full_feature_names=full_feature_names,
-            )
-
             try:
-                yield query
+                yield build_point_in_time_query(
+                    query_context,
+                    left_table_query_string=df_query,
+                    entity_df_event_timestamp_col=entity_df_event_timestamp_col,
+                    entity_df_columns=entity_schema.keys(),
+                    query_template=MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN,
+                    full_feature_names=full_feature_names,
+                )
             finally:
-                with _get_conn(config.offline_store) as conn, conn.cursor() as cur:
-                    cur.execute(
-                        sql.SQL(
-                            """
-                            DROP TABLE IF EXISTS {};
-                            """
-                        ).format(sql.Identifier(table_name)),
-                    )
+                if table_name:
+                    with _get_conn(config.offline_store) as conn, conn.cursor() as cur:
+                        cur.execute(
+                            sql.SQL(
+                                """
+                                DROP TABLE IF EXISTS {};
+                                """
+                            ).format(sql.Identifier(table_name)),
+                        )
 
         return PostgreSQLRetrievalJob(
             query=query_generator,
@@ -142,17 +161,6 @@ class PostgreSQLOfflineStore(OfflineStore):
         )
 
 
-def _df_to_table(
-    config: PostgreSQLConfig, entity_df: Union[pd.DataFrame, str], table_name: str
-) -> Dict[str, str]:
-    if isinstance(entity_df, pd.DataFrame):
-        return df_to_postgres_table(config, entity_df, table_name)
-    elif isinstance(entity_df, str):
-        return sql_to_postgres_table(config, entity_df, table_name)
-    else:
-        raise TypeError(entity_df)
-
-
 class PostgreSQLRetrievalJob(RetrievalJob):
     def __init__(
         self,
@@ -172,7 +180,6 @@ class PostgreSQLRetrievalJob(RetrievalJob):
 
             self._query_generator = query_generator
         self.config = config
-        self.connection = _get_conn(self.config.offline_store)
         self._full_feature_names = full_feature_names
         self._on_demand_feature_views = on_demand_feature_views
 
@@ -194,13 +201,13 @@ class PostgreSQLRetrievalJob(RetrievalJob):
 
     def _to_arrow_internal(self) -> pa.Table:
         with self._query_generator() as query:
-            with self.connection as conn, conn.cursor() as cur:
+            with _get_conn(self.config.offline_store) as conn, conn.cursor() as cur:
+                conn.set_session(readonly=True)
                 cur.execute(query)
                 fields = [
                     (c.name, pg_type_code_to_arrow(c.type_code))
                     for c in cur.description
                 ]
-                print(fields)
                 data = cur.fetchall()
                 schema = pa.schema(fields)
                 # TODO: Fix...
@@ -217,8 +224,42 @@ class PostgreSQLRetrievalJob(RetrievalJob):
 
 
 def _append_alias(field_names: List[str], alias: str) -> List[str]:
-    return [f"{alias}.{field_name}" for field_name in field_names]
+    return [f"{alias}.\"{field_name}\"" for field_name in field_names]
 
+def build_point_in_time_query(
+    feature_view_query_contexts: List[offline_utils.FeatureViewQueryContext],
+    left_table_query_string: str,
+    entity_df_event_timestamp_col: str,
+    entity_df_columns: KeysView[str],
+    query_template: str,
+    full_feature_names: bool = False,
+) -> str:
+    """Build point-in-time query between each feature view table and the entity dataframe for Bigquery and Redshift"""
+    template = Environment(loader=BaseLoader()).from_string(source=query_template)
+
+    final_output_feature_names = list(entity_df_columns)
+    final_output_feature_names.extend(
+        [
+            (f"{fv.name}__{feature}" if full_feature_names else feature)
+            for fv in feature_view_query_contexts
+            for feature in fv.features
+        ]
+    )
+
+    # Add additional fields to dict
+    template_context = {
+        "left_table_query_string": left_table_query_string,
+        "entity_df_event_timestamp_col": entity_df_event_timestamp_col,
+        "unique_entity_keys": set(
+            [entity for fv in feature_view_query_contexts for entity in fv.entities]
+        ),
+        "featureviews": [asdict(context) for context in feature_view_query_contexts],
+        "full_feature_names": full_feature_names,
+        "final_output_feature_names": final_output_feature_names,
+    }
+
+    query = template.render(template_context)
+    return query
 
 # Copied from the Feast Redshift offline store implementation
 # Note: Keep this in sync with sdk/python/feast/infra/offline_stores/redshift.py:
@@ -285,9 +326,9 @@ WITH entity_dataframe AS (
         {{ featureview.created_timestamp_column ~ ' as created_timestamp,' if featureview.created_timestamp_column else '' }}
         {{ featureview.entity_selections | join(', ')}}{% if featureview.entity_selections %},{% else %}{% endif %}
         {% for feature in featureview.features %}
-            {{ feature }} as {% if full_feature_names %}{{ featureview.name }}__{{feature}}{% else %}{{ feature }}{% endif %}{% if loop.last %}{% else %}, {% endif %}
+            "{{ feature }}" as "{% if full_feature_names %}{{ featureview.name }}__{{feature}}{% else %}{{ feature }}{% endif %}"{% if loop.last %}{% else %}, {% endif %}
         {% endfor %}
-    FROM {{ featureview.table_subquery }} sub
+    FROM {{ featureview.table_subquery }} AS sub
     WHERE {{ featureview.event_timestamp_column }} <= (SELECT MAX(entity_timestamp) FROM entity_dataframe)
     {% if featureview.ttl == 0 %}{% else %}
     AND {{ featureview.event_timestamp_column }} >= (SELECT MIN(entity_timestamp) FROM entity_dataframe) - {{ featureview.ttl }} * interval '1' second
@@ -336,19 +377,23 @@ WITH entity_dataframe AS (
 */
 {{ featureview.name }}__latest AS (
     SELECT
-        {{featureview.name}}__entity_row_unique_id,
-        MAX(event_timestamp) AS event_timestamp
+        event_timestamp,
+        {% if featureview.created_timestamp_column %}created_timestamp,{% endif %}
+        {{featureview.name}}__entity_row_unique_id
+    FROM
+    (
+        SELECT *,
+            ROW_NUMBER() OVER(
+                PARTITION BY {{featureview.name}}__entity_row_unique_id
+                ORDER BY event_timestamp DESC{% if featureview.created_timestamp_column %},created_timestamp DESC{% endif %}
+            ) AS row_number
+        FROM {{ featureview.name }}__base
         {% if featureview.created_timestamp_column %}
-            ,MAX(created_timestamp) AS created_timestamp
+            INNER JOIN {{ featureview.name }}__dedup
+            USING ({{featureview.name}}__entity_row_unique_id, event_timestamp, created_timestamp)
         {% endif %}
-
-    FROM {{ featureview.name }}__base
-    {% if featureview.created_timestamp_column %}
-        INNER JOIN {{ featureview.name }}__dedup
-        USING ({{featureview.name}}__entity_row_unique_id, event_timestamp, created_timestamp)
-    {% endif %}
-
-    GROUP BY {{featureview.name}}__entity_row_unique_id
+    ) AS sub
+    WHERE row_number = 1
 ),
 
 /*
@@ -366,7 +411,7 @@ WITH entity_dataframe AS (
             ,created_timestamp
         {% endif %}
     )
-) {% if loop.last %}{% else %}, {% endif %}
+){% if loop.last %}{% else %}, {% endif %}
 
 
 {% endfor %}
@@ -375,16 +420,16 @@ WITH entity_dataframe AS (
  The entity_dataframe dataset being our source of truth here.
  */
 
-SELECT *
+SELECT "{{ final_output_feature_names | join('", "')}}"
 FROM entity_dataframe
 {% for featureview in featureviews %}
 LEFT JOIN (
     SELECT
-        {{featureview.name}}__entity_row_unique_id
+        "{{featureview.name}}__entity_row_unique_id"
         {% for feature in featureview.features %}
-            ,{% if full_feature_names %}{{ featureview.name }}__{{feature}}{% else %}{{ feature }}{% endif %}
+            ,"{% if full_feature_names %}{{ featureview.name }}__{{feature}}{% else %}{{ feature }}{% endif %}"
         {% endfor %}
     FROM {{ featureview.name }}__cleaned
-) {{featureview.name}} USING ({{featureview.name}}__entity_row_unique_id)
+) AS "{{featureview.name}}" USING ({{featureview.name}}__entity_row_unique_id)
 {% endfor %}
 """
